{ Andre's Program Engine (APE) Components

  Autor: Demidov Andre (E-mail: andre@lion.tu-chel.ac.ru)
  Version 1.0
  (c) 1997
  
  This unit make easy a programming of DB-apps, based on
  table-form conception (a-la Clarion and others).

  This unit contain three components
  1. TApeMainForm, what transform main form in MDI Form and
           keep names of MDI Child.
  2. TApeForm, what transform current form in MDI Child with
           new possibilities.
  3. TApeDialog, what permit to execute other MDI Child with
           TApeForm component in pseudo-Modal mode. The owner
           must contain TApeForm xor TApeMainForm component too.

  At start program (project source is the best place for this)
  you must register your form classes (use RegisterClass or
  RegisterClasses). 
  
  Each of ApeDialogs of the main form correspond to a
  process. This process may be run once only. If the process 
  already ran then the last active form of process will select.
  
  Also each of ApeDialogs of the child form can give birth to one
  form. If form already exists then it will select. If ExecuteMode of
  the new form is NextLevel or SelectValue then all forms of this 
  process with smaller level will disable. But, when this form closed,
  they will enable and the parent form will select.

  --TApeMainForm
  must be sole in project and placed on main form of program

  --TApeForm
  must be sole in the form

  Properties:
      FormMode:ApeFormMode;
          can use in design time only
          to indicate form style
             FormWindow - for editing one DB record
                          the form can't change its sizes
             BrowseWindow - for browsing the list of records
                          the form can change its sizes
      SelectResult:string;
          to return selected value after selection

   Metods:
       function GetParentForm:TForm;
           to get parent form

   Events:
      onFirstShow(Sender:TObject;
           PMode:ApeActivateFormMode;
           PObject:TObject;
           PValue,
           PSelect:string);
         run at first show event for this form
      PMode can has one from two values
           SelectObject (if need selection object)
           and CreateObject (in other cases)
      PObject can be TTable or TStringList with many parameters
      PValue use for transfering work mode:
         'E' - edit
         'I' - insert
         'F1' - use the filter N1 etc.
      PSelect is ID selected record

  --TApeDialog
  Properties:
      property FormClassName:string;
         class of form what must run
      property withValue:string;
         additional parameter for transfering work mode
          'E' - edit
          'I' - insert
          'F1' - use the filter N1 etc.
      property withSelect:string;
          ID selected record
      property withObject:TComponent;
          nil or TTable or TStringList with many parameters
      property ExecuteMode:ApeCreateFormMode;
          execution mode can be one from three:
            ThisLevel enable selection this form
            NextLevel disable selection this form while child run
            SelectValue is same but allow transfer selected value
      property ProcessID:integer,readonly;

Metods:
      function GetChildForm:TForm;
         to get child form, return nil if not running
      procedure Execute;
         execute form with FormClassName

Events:
  OnSelect(Sender:TObject;
           SelectedValue:string);
     run when child form closed and SelectResult be set in
         any new value
}

unit Ape;

interface
uses Forms, Classes, sysUtils;

type ApeCreateFormMode=(ThisLevel,
                        NextLevel,   
                        SelectValue);

     ApeActivateFormMode=(NoneAction,
                          SelectObject,
                          WaitObject,
                          CreateObject);

     ApeFormMode=(FormWindow,
                 BrowseWindow);

    TApeSelectProc=procedure(Sender:TObject; SelectedValue:string) of object;
    TApeShowProc=procedure(Sender:TObject; PMode:ApeActivateFormMode;
         PObject:TObject; PValue,PSelect:string) of object;

    TApeDialog=class(TComponent)
    private
      FFormClassName:string;
      FObjectValue:TComponent;
      FParamValue:string;
      FSelectValue:string;
      FSelect:TApeSelectProc;
      FExecuteMode:ApeCreateFormMode;
      FprocessID:integer;
    public
      constructor Create(AOwner:TComponent); override;
      function GetChildForm:TForm;
      procedure Execute;
      property ProcessID:integer read FprocessID write FprocessID;
    published
      property FormClassName:string read FFormClassName write FFormClassName;
      property withValue:string read FParamValue write FParamValue;
      property withSelect:string read FSelectValue write FSelectValue;
      property withObject:TComponent read FObjectValue write FObjectValue;
      property ExecuteMode:ApeCreateFormMode read FExecuteMode write FExecuteMode;
      property OnSelect:TApeSelectProc read FSelect write FSelect;
    end;

    TApeForm=class(TComponent)
    private
      oldActivateProc:TNotifyEvent;
      OldShowProc:TNotifyEvent;
      oldCloseProc:TCloseEvent;
      FFirstShow:TApeShowProc;
      FFormMode:ApeFormMode;
      FSelectResult:string;
      procedure SetFormMode(FMode:ApeFormMode);
      procedure SetSelectResult(FResult:string);
      procedure ActivateProc(Sender:TObject);
      procedure CloseProc(Sender:TObject; var Action: TCloseAction);
      procedure ShowProc(Sender:TObject);
    public
      constructor Create(AOwner:TComponent); override;
      destructor Destroy; override;
      procedure Loaded; override;
      function GetParentForm:TForm;
      property SelectResult:string read FSelectResult write SetSelectResult;
    published
      property FormMode:ApeFormMode read FFormMode write setFormMode;
      property onFirstShow:TApeShowProc read FFirstShow write FFirstShow;
    end;

    TApeMainForm=class(TComponent)
    private
      FFormNameList: TStrings;
    public
      constructor Create(AOwner:TComponent); override;
      destructor Destroy; override;
      procedure AddFormName(val: string);
      procedure SetFormNameList(val:TStrings);
    published
      property FormNameList: TStrings read FFormNameList write SetFormNameList;
    end;

procedure ApeInit;
procedure ApeCreateProcess(PClass:TFormClass; PID:integer;
          var PForm; PValue:string);
procedure ApeSelectProcess(PID:integer);
function ApeProcessRunning(PID:integer):Boolean;
procedure ApeCreateForm(PClass:TFormClass; PFormP:TForm;
          var PForm; PMode:ApeCreateFormMode;
          PObject:TObject; PValue:string; PSelect:string);
procedure ApeCreateForm2(PClass:TFormClass; PFormP:TForm;
          var PForm; PMode:ApeCreateFormMode;
          PObject1,PObject2:TObject; PValue:string; PSelect:string);
procedure ApeCloseForm(PForm:TForm);
procedure ApeSelectForm(PForm:TForm);
procedure ApeGetMode(PForm:TForm; var PMode:ApeActivateFormMode;
                   var PObject:TObject; var PValue:string; var PSelect:string);
function ApeGetParent(PForm:TForm):TForm;
function ApeGetChild(PForm:TForm;PClass:TFormClass):TForm;
function ApeGetProcessChild(PID:integer):TForm;
procedure ApeSetSelect(PForm:TForm; PSelect:string);

var
    ApeMainForm: TApeMainForm;

implementation

const MaxNForm=200;
      MaxNProcess=100;

type ProcessInfoRecord=
    record
       PClass:TFormClass;
       PForm:TForm;
       PFormParent:TForm;
       PLevel:integer;
       PID:integer;
       PMode:ApeActivateFormMode;
       PObject:TObject;
       PValue:string[32];
       PSelect:string[32];
    end;

var ProcessInfo:array [1..200] of ProcessInfoRecord;
    CurrentInfo:array [1..100] of TForm;
    NForm:integer;
    ApeMainFormCount: integer;

procedure ApeInit;
begin
  NForm:=0;
  ApeMainForm:=nil;
  ApeMainFormCount:=0;
end;

{ TApeMainForm }

procedure TApeMainForm.SetFormNameList(val:TStrings);
begin
  FFormNameList.Assign(val);
end;

procedure TApeMainForm.AddFormName(val:string);
begin
  if (FFormNameList.IndexOf(val)>-1) then exit;
  FFormNameList.Add(val);
end;

constructor TApeMainForm.Create(AOwner:Tcomponent);
begin
  inc(ApeMainFormCount);
  if (AOwner<>nil) then
    with AOwner as TForm do
    begin
        if (ApeMainForm<>nil) then
        Raise Exception.Create('TApeMainForm must be sole in the project!');
    end;

  inherited Create(Aowner);
  if (csDesigning in ComponentState)  and (AOwner<>nil) then
    with AOwner as TForm do
    begin
      FormStyle:=fsMDIForm;
    end;
  ApeMainForm:=self;
  FFormNameList:= TStringList.Create;
end;

destructor TApeMainForm.Destroy;
begin
  FFormNameList.Free;
  inherited Destroy;
  dec(ApeMainFormCount);
  if (ApeMainFormCount=0) then
    ApeMainForm:=nil;
end;

{ TApeForm }

constructor TApeForm.Create(AOwner:Tcomponent);
var
   i: integer;
begin
  if (AOwner<>nil) then
    with AOwner as TForm do
    begin
    for I := 0 to ComponentCount -1 do
     if (Components[I] is TApeForm) or (Components[I] is TApeMainForm) then
        Raise Exception.Create('TApeForm must be sole on the form!');
    end;

  inherited Create(Aowner);
  oldActivateProc:=nil;
  OldShowProc:=nil;
  oldCloseProc:=nil;
  FFormMode:=FormWindow;
  if (csDesigning in ComponentState)  and (AOwner<>nil) then
  begin
    with AOwner as TForm do
    begin
      if (ApeMainForm<>nil) then
         ApeMainForm.AddFormName(ClassName);
      FormStyle:=fsMDIChild;
      Position:=poDefaultPosOnly;
      if FFormMode=FormWindow then
      begin
        BorderIcons:=[biSystemMenu,biMinimize];
        BorderStyle:=bsSingle;
      end;
    end;
  end;
end;

destructor TApeForm.Destroy;
begin
  if not (csDesigning in ComponentState)  and (Owner<>nil) then
  begin
    with Owner as TForm do
    begin
      OnShow:=oldShowProc;
      OnActivate:=oldActivateProc;
      OnClose:=oldCloseProc;
    end;
  end;
  inherited Destroy;
end;

function TApeForm.GetParentForm:TForm;
begin
  Result:=ApeGetParent(TForm(Owner));
end;

procedure TApeForm.Loaded;
begin
  inherited Loaded;
  if not (csDesigning in ComponentState)  and (Owner<>nil) then
  begin
    with Owner as TForm do
    begin
      oldShowProc:=onShow;
      oldActivateProc:=onActivate;
      oldCloseProc:=onClose;
      onShow:=ShowProc;
      onActivate:=ActivateProc;
      onClose:=CloseProc;
    end;
  end
  else
  if (csDesigning in ComponentState)  and (Owner<>nil) then
  begin
    with Owner as TForm do
    begin
      FormStyle:=fsMDIChild;
      Position:=poDefaultPosOnly;
      if FFormMode=FormWindow then
      begin
        BorderIcons:=[biSystemMenu,biMinimize];
        BorderStyle:=bsSingle;
      end;
    end;
  end;
end;

procedure TApeForm.SetSelectResult(FResult:string);
begin
  ApeSetSelect(TForm(Owner),FResult);
  FSelectResult:=FResult;
end;

procedure TApeForm.SetFormMode(FMode:ApeFormMode);
begin
  if FFormMode<>FMode then
  begin
    FFormMode:=FMode;
  if (csDesigning in ComponentState)  and (Owner<>nil) then
  begin
    with Owner as TForm do
    begin
      FormStyle:=fsMDIChild;
      Position:=poDefaultPosOnly;
      if FFormMode=FormWindow then
      begin
        BorderIcons:=[biSystemMenu,biMinimize];
        BorderStyle:=bsSingle;
      end
      else
      begin
        BorderIcons:=[biSystemMenu,biMinimize,biMaximize];
        BorderStyle:=bsSizeable;
      end;
    end;
  end;
  end;
end;

procedure TApeForm.ActivateProc(Sender:TObject);
var PMode:ApeActivateFormMode;
    PObject:TObject;
    PValue:string;
    PSelect:string;
    SProc:TApeSelectProc;
begin
  ApeGetMode(TForm(Owner),Pmode,Pobject,PValue,Pselect);
  if Pmode=waitObject then
  begin
    if Pobject is TApeDialog then
    begin
      with PObject as TApeDialog do
      begin
        sProc:=onSelect;
        if assigned(sProc) then
          onSelect(PObject,Pselect);
      end;
    end;
  end;
  ApeSelectForm(TForm(Owner));
end;

procedure TApeForm.CloseProc(Sender:TObject; var Action: TCloseAction);
begin
  if assigned(oldCloseProc) then oldCloseProc(Sender,Action);
  Action:=caFree;
  ApeCloseForm(Tform(owner));
end;

procedure TApeForm.ShowProc(Sender:TObject);
var PMode:ApeActivateFormMode;
    PObject:TObject;
    PValue:string;
    PSelect:string;
begin
  ApeGetMode(TForm(Owner),Pmode,Pobject,PValue,Pselect);
  if Pmode<>noneAction then
    if assigned(FFirstShow) then
      FFirstShow(sender,PMode,Pobject,Pvalue,Pselect);
  if assigned(oldShowProc) then
    oldShowProc(sender);
end;

{ TApeDialog }

constructor TApeDialog.Create(AOwner:TComponent);
var i,PID,maxPID:integer;
   cmpt:TComponent;
begin
  inherited Create(Aowner);
  FFormClassName:='';
  FObjectValue:=nil;
  FParamValue:='';
  FSelectValue:='';
  FSelect:=nil;
  FExecuteMode:=SelectValue;
  FProcessId:=0;
  maxPID:=0;
  for i:=0 to (AOwner as TForm).ComponentCount-1 do
  begin
    cmpt:=(AOwner as TForm).Components[i];
    if cmpt is TApeDialog then
    begin
      PID:=(cmpt as TApeDialog).FprocessID;
      if PID>maxPID then
        maxPID:=PID;
    end;
  end;
  FProcessId:=maxPID+1;
end;

procedure TApeDialog.Execute;
var PForm:TForm;
    PClass:TFormClass;
begin
  PClass:=TFormClass(GetClass(FFormClassName));
  if PClass<>nil then
  begin
  if (Owner as TForm).FormStyle=fsMDIChild then
  begin
    ApeCreateForm2(PClass,
       TForm(Owner),PForm,FExecuteMode,Self,FobjectValue,
          FParamValue,FSelectValue);
  end
  else
  begin
    ApeCreateProcess(PClass, ProcessID,
          PForm, FParamValue);
  end;
  end;
end;

function TApeDialog.GetChildForm:TForm;
var PClass:TFormClass;
begin
  Result:=nil;
  if (Owner as TForm).FormStyle=fsMDIChild then
  begin
    Pclass:=TFormClass(GetClass(FFormClassName));
    if Pclass<>nil then
      Result:=ApeGetChild(Tform(Owner), Pclass);
  end
  else
    Result:=ApeGetProcessChild(ProcessID);
end;

function findInfo(PForm:TForm):integer;
var i:integer;
begin
  Result:=0;
  for I:=1 to NForm do
  begin
    if PForm=ProcessInfo[i].Pform then
    begin
      Result:=i;
      break;
    end;
  end;
end;

procedure ApeCreateProcess(PClass:TFormClass; PID:integer;
          var PForm; Pvalue:string);
var i:integer;
begin
  if (PID<=0) or (PID>MaxNProcess) then
  begin
    TForm(PForm):=nil;
  end
  else if ApeProcessRunning(PID) then
  begin
    ApeSelectProcess(PID);
    TForm(PForm):=nil;
  end
  else
  begin
    i:=findInfo(nil);
    if i=0 then
    begin
      NForm:=NForm+1;
      i:=NForm;
    end;
    ProcessInfo[i].PClass:=PClass;
    ProcessInfo[i].PValue:=PValue;
    ProcessInfo[i].PSelect:='';
    ProcessInfo[i].PObject:=nil;
    ProcessInfo[i].PMode:=CreateObject;
    ProcessInfo[i].PID:=PID;
    ProcessInfo[i].PLevel:=1;
    ProcessInfo[i].PFormParent:=nil;
    Application.CreateForm(PClass,ProcessInfo[i].PForm);
    TForm(PForm):=ProcessInfo[i].PForm;
    TForm(PForm).Show;
  end;
end;

procedure ApeSelectProcess(PID:integer);
var i:integer;
begin
  if (PID<=0) or (PID>MaxNProcess) then
    Exit;
  if CurrentInfo[PID]<>nil then
  begin
    i:=findInfo(CurrentInfo[PID]);
    if i>0 then
    begin
      with CurrentInfo[PID] do
      begin
        if windowState=wsMinimized then
          windowState:=wsNormal;
        Show;
      end;
    end;
  end;
end;

function ApeProcessRunning(PID:integer):Boolean;
var i:integer;
begin
  Result:=False;
  if PID=0 then
  begin
    for I:=1 to NForm do
    begin
      if (ProcessInfo[i].PForm<>nil) then
      begin
        Result:=True;
        Exit;
      end;
    end;
  end;
  if (PID<=0) or (PID>MaxNProcess) then
    Exit;
  for I:=1 to NForm do
  begin
    if (ProcessInfo[i].PForm<>nil) and
       (PID=ProcessInfo[i].PID) then
    begin
      Result:=True;
      Break;
    end;
  end;
end;

function ApeGetProcessChild(PID:integer):TForm;
var i:integer;
begin
  Result:=nil;
  if (PID<=0) or (PID>MaxNProcess) then
    Exit;
  for I:=1 to NForm do
  begin
    if (ProcessInfo[i].PForm<>nil) and
       (ProcessInfo[i].PFormParent=nil) and
       (PID=ProcessInfo[i].PID) then
    begin
      Result:=ProcessInfo[i].PForm;
      Break;
    end;
  end;
end;

procedure ApeCreateForm(PClass:TFormClass; PFormP:TForm;
          var PForm; PMode:ApeCreateFormMode;
          PObject:TObject; PValue:string; Pselect:string);
begin
  if Pmode=SelectValue then
    ApeCreateForm2(Pclass,PFormP,PForm,PMode,PObject,nil,PValue,Pselect)
  else
    ApeCreateForm2(Pclass,PFormP,PForm,PMode,nil,PObject,PValue,Pselect);
end;

procedure ApeCreateForm2(PClass:TFormClass; PFormP:TForm;
          var PForm; PMode:ApeCreateFormMode;
          PObject1, PObject2:TObject; PValue:string; Pselect:string);
var p,i,lvl,j:integer;
begin
  for i:=1 to Nform do
    if (ProcessInfo[i].Pform<>nil) and
       (ProcessInfo[i].PformParent=PFormP) and
       (ProcessInfo[i].PClass=PClass) then
    begin
      TForm(PForm):=ProcessInfo[i].Pform;
      with ProcessInfo[i].Pform do
      begin
        if windowState=wsMinimized then
          windowState:=wsNormal;
        Show;
      end;
      Exit;
    end;
  p:=findInfo(PFormP);
  if p=0 then
    TForm(PForm):=nil
  else
  begin
    i:=findInfo(nil);
    if i=0 then
    begin
      NForm:=NForm+1;
      i:=NForm;
    end;
    ProcessInfo[i].PForm:=nil;
    ProcessInfo[i].PValue:=PValue;
    ProcessInfo[i].PSelect:=PSelect;
    ProcessInfo[i].PClass:=PClass;
    ProcessInfo[p].PSelect:='';
    ProcessInfo[p].PMode:=NoneAction;
    ProcessInfo[p].PObject:=PObject1;
    ProcessInfo[i].PObject:=Pobject2;
    if Pmode=SelectValue then
      begin
        ProcessInfo[i].PMode:=SelectObject;
      end
    else
      begin
        ProcessInfo[i].PMode:=CreateObject;
      end;
    ProcessInfo[i].PID:=ProcessInfo[p].PID;
    ProcessInfo[i].PFormParent:=PFormP;
    if Pmode=ThisLevel then
      ProcessInfo[i].PLevel:=ProcessInfo[p].Plevel
    else
    begin
      ProcessInfo[i].PLevel:=ProcessInfo[p].Plevel+1;
      for j:=1 to NForm do
      begin
        if (ProcessInfo[j].PLevel=ProcessInfo[p].Plevel) and
           (ProcessInfo[j].PID=ProcessInfo[p].PID) and
          (ProcessInfo[j].PForm<>nil) then
          ProcessInfo[j].PForm.Enabled:=false;
      end;
    end;
    Application.CreateForm(PClass,ProcessInfo[i].PForm);
    TForm(PForm):=ProcessInfo[i].PForm;
    TForm(PForm).Show;
  end;
end;

procedure ApeCloseForm(PForm:TForm);
var i,j,p:integer;
begin
  i:=findInfo(PForm);
  if i>0 then
  begin
    for j:=1 to Nform do
    begin
      if (ProcessInfo[j].PForm<>nil) and
         (ProcessInfo[j].PFormParent=PForm) then
        ProcessInfo[j].PForm.Close;
    end;
    ProcessInfo[i].PForm:=nil;
    if ProcessInfo[i].PformParent<>nil then
    begin
    p:=findInfo(ProcessInfo[i].PformParent);
    if p>0 then
    begin
      if ProcessInfo[p].Plevel<>ProcessInfo[i].Plevel then
      begin
        {    }
        for j:=1 to NForm do
        begin
          if (ProcessInfo[j].PLevel=ProcessInfo[p].Plevel) and
             (ProcessInfo[j].PID=ProcessInfo[p].PID) and
             (ProcessInfo[j].PForm<>nil) then
            ProcessInfo[j].PForm.Enabled:=True;
        end;
      end;
      with ProcessInfo[p].Pform do
      begin
        if windowState=wsMinimized then
          windowState:=wsNormal;
        Show;
      end;
    end;
    end
    else if ProcessInfo[i].Plevel=1 then
    begin
      CurrentInfo[ProcessInfo[i].PID]:=nil;
    end;
  end;
end;

procedure ApeSelectForm(PForm:TForm);
var i:integer;
begin
  i:=findInfo(PForm);
  if i>0 then
  begin
    CurrentInfo[ProcessInfo[i].PID]:=PForm;
    ProcessInfo[i].Pmode:=NoneAction;
    ProcessInfo[i].PValue:='';
    ProcessInfo[i].PSelect:='';
  end;
end;

procedure ApeGetMode(PForm:TForm; var PMode:ApeActivateFormMode;
                   var PObject:TObject; var PValue:string; var PSelect:string);
var i:integer;
begin
  i:=findInfo(PForm);
  if i>0 then
  begin
    Pmode:=ProcessInfo[i].Pmode;
    PObject:=ProcessInfo[i].Pobject;
    Pvalue:=ProcessInfo[i].Pvalue;
    PSelect:=ProcessInfo[i].PSelect;
  end
  else
  begin
    Pmode:=NoneAction;
    PObject:=nil;
    Pvalue:='';
    PSelect:='';
  end;
end;

procedure ApeSetSelect(PForm:TForm; PSelect:string);
var i,p:integer;
begin
  i:=findInfo(PForm);
  if i>0 then
  begin
     p:=findInfo(ProcessInfo[i].PformParent);
     if p>0 then
     begin
       ProcessInfo[p].PSelect:=PSelect;
       ProcessInfo[p].Pmode:=WaitObject;
     end;
  end;
end;

function ApeGetParent(PForm:TForm):TForm;
var i:integer;
begin
  I:=findinfo(PForm);
  Result:=nil;
  if i>0 then
  begin
    if findinfo(ProcessInfo[i].PformParent)>0 then
      Result:=ProcessInfo[i].PformParent;
  end;
end;

function ApeGetChild(PForm:TForm; PClass:TFormClass):TForm;
var i:integer;
begin
  Result:=nil;
  for i:=1 to Nform do
    if (ProcessInfo[i].Pform<>nil) and
       (ProcessInfo[i].PformParent=PForm) and
       (ProcessInfo[i].PClass=PClass) then
    begin
      Result:=ProcessInfo[i].Pform;
      Break;
    end;
end;

initialization
  ApeInit
end.
